import jax.numpy as jnp
import jax.random as jr
import jax.tree_util as jtu
from jaxtyping import ArrayLike

from ...core import (
    _cdf_impl,
    _cf_impl,
    _entropy_impl,
    _intialize_params_tree,
    _kurtosis_impl,
    _logcdf_impl,
    _logpdf_impl,
    _mean_impl,
    _mgf_impl,
    _params_impl,
    _pdf_impl,
    _quantile_impl,
    _rand_impl,
    _sf_impl,
    _skewness_impl,
    _standard_dev_impl,
    _support_impl,
    _variance_impl,
    DTypeLikeFloat,
    KeyArray,
    PyTreeVar,
    Shape,
)
from ...dist_math.exp import (
    exp_cdf,
    exp_cf,
    exp_logcdf,
    exp_logpdf,
    exp_mgf,
    exp_pdf,
    exp_ppf,
    exp_sf,
)
from ...random_utils import split_tree
from ...tree_utils import tree_map_dist_at
from .._base import ContinuousUnivariateDistribution


class Exponential(ContinuousUnivariateDistribution):
    """Exponential distribution.

    Args:
        rate (PyTree): Rate parameter.
        dtype (jax.numpy.dtype): dtype of the distribution, default jnp.float_.
        use_batch (bool): Whether to use with vmap. Default False.

    Examples:
        >>> import jax.numpy as jnp
        >>> from fenbux import Exponential, logpdf
        >>> dist = Exponential(1.0)
        >>> logpdf(dist, jnp.ones((10, )))
    """

    rate: PyTreeVar

    def __init__(
        self,
        rate: PyTreeVar = 1.0,
        dtype=jnp.float_,
        use_batch=False,
    ):
        self.rate = _intialize_params_tree(rate, use_batch=use_batch, dtype=dtype)


@_params_impl.dispatch
def _params(d: Exponential):
    return (d.rate,)


@_support_impl.dispatch
def _support(d: Exponential):
    return jtu.tree_map(lambda r: jnp.zeros_like(r), d.rate), jtu.tree_map(
        lambda r: jnp.full_like(r, jnp.inf), d.rate
    )


@_mean_impl.dispatch
def _mean(d: Exponential):
    return jtu.tree_map(lambda x: 1.0 / x, d.rate)


@_variance_impl.dispatch
def _variance(d: Exponential):
    return jtu.tree_map(lambda x: 1.0 / x**2, d.rate)


@_standard_dev_impl.dispatch
def _standard_dev(d: Exponential):
    return jtu.tree_map(lambda x: 1.0 / x, d.rate)


@_skewness_impl.dispatch
def _skewness(d: Exponential):
    return jtu.tree_map(lambda x: 2.0, d.rate)


@_kurtosis_impl.dispatch
def _kurtosis(d: Exponential):
    return jtu.tree_map(lambda x: 6.0, d.rate)


@_support_impl.dispatch
def _support(d: Exponential):
    return jtu.tree_map(lambda x: (0.0, jnp.inf), d.rate)


@_entropy_impl.dispatch
def _entropy(d: Exponential):
    return jtu.tree_map(lambda x: 1.0 - jnp.log(x), d.rate)


@_logpdf_impl.dispatch
def _logpdf(d: Exponential, x: ArrayLike):
    return tree_map_dist_at(exp_logpdf, d, x)


@_pdf_impl.dispatch
def _pdf(d: Exponential, x: ArrayLike):
    return tree_map_dist_at(exp_pdf, d, x)


@_logcdf_impl.dispatch
def _logcdf(d: Exponential, x: ArrayLike):
    return tree_map_dist_at(exp_logcdf, d, x)


@_cdf_impl.dispatch
def _cdf(d: Exponential, x: ArrayLike):
    return tree_map_dist_at(exp_cdf, d, x)


@_quantile_impl.dispatch
def _quantile(d: Exponential, x: ArrayLike):
    return tree_map_dist_at(exp_ppf, d, x)


@_sf_impl.dispatch
def _sf(d: Exponential, x: ArrayLike):
    return tree_map_dist_at(exp_sf, d, x)


@_mgf_impl.dispatch
def _mgf(d: Exponential, t: ArrayLike):
    return tree_map_dist_at(exp_mgf, d, t)


@_cf_impl.dispatch
def _cf(d: Exponential, t: ArrayLike):
    return tree_map_dist_at(exp_cf, d, t)


@_rand_impl.dispatch
def _rand(
    d: Exponential, key: KeyArray, shape: Shape = (), dtype: DTypeLikeFloat = float
):
    _key_tree = split_tree(key, d.rate)
    return jtu.tree_map(
        lambda r, k: jr.exponential(k, shape, dtype) / r, d.rate, _key_tree
    )



import numpy as np
from scipy import stats

from fenbux.scipy_stats import bernoulli, binom, chi2, gamma, norm, poisson, t, uniform

from .helpers import shaped_allclose


def test_create_dists():
    assert norm(1.0, 2.0).mean() == stats.norm(1.0, 2.0).mean()
    assert gamma(1.0, 2.0).mean() == stats.gamma(1.0, 2.0).mean()
    assert chi2(1.0).mean() == stats.chi2(1.0).mean()
    assert poisson(1.0).mean() == stats.poisson(1.0).mean()
    assert uniform(1.0, 2.0).mean() == stats.uniform(1.0, 2.0).mean()
    assert t(df=2.0).mean() == stats.t(df=2.0).mean()
    assert binom(1.0, 0.2).mean() == stats.binom(1.0, 0.2).mean()
    assert bernoulli(0.2).mean() == stats.bernoulli(0.2).mean()


def test_create_tree_dists():
    assert norm({"a": 1.0, "b": 2.0}, {"a": 2.0, "b": 3.0}).mean() == {
        "a": stats.norm(1.0, 1.0).mean(),
        "b": stats.norm(2.0, 2.0).mean(),
    }
    assert gamma({"a": 1.0, "b": 2.0}, {"a": 2.0, "b": 3.0}).mean() == {
        "a": stats.gamma(1.0, 2.0).mean(),
        "b": stats.gamma(2.0, 3.0).mean(),
    }
    assert chi2({"a": 1.0, "b": 2.0}).mean() == {
        "a": stats.chi2(1.0).mean(),
        "b": stats.chi2(2.0).mean(),
    }
    assert poisson({"a": 1.0, "b": 2.0}).mean() == {
        "a": stats.poisson(1.0).mean(),
        "b": stats.poisson(2.0).mean(),
    }


def test_tree_io():
    normal = norm({"a": 1.0, "b": 2.0}, {"a": 2.0, "b": 3.0})
    x = np.random.uniform(size=10000)
    assert shaped_allclose(
        normal.logpdf(x=x, flat_kwargnames=["x"]),
        {"a": stats.norm(1.0, 2.0).logpdf(x), "b": stats.norm(2.0, 3.0).logpdf(x)},
    )
    assert shaped_allclose(
        normal.pdf(x=x, flat_kwargnames=["x"]),
        {"a": stats.norm(1.0, 2.0).pdf(x), "b": stats.norm(2.0, 3.0).pdf(x)},
    )
    assert shaped_allclose(
        normal.cdf(x=x, flat_kwargnames=["x"]),
        {"a": stats.norm(1.0, 2.0).cdf(x), "b": stats.norm(2.0, 3.0).cdf(x)},
    )
    assert shaped_allclose(
        normal.ppf(q=x, flat_kwargnames=["q"]),
        {"a": stats.norm(1.0, 2.0).ppf(x), "b": stats.norm(2.0, 3.0).ppf(x)},
    )



from jax import numpy as jnp
from jax.scipy.special import betainc, gammaln
from jaxtyping import Array
from tensorflow_probability.substrates.jax.math import betaincinv


def t_logpdf(x, df) -> Array:
    return (
        gammaln((df + 1) / 2)
        - gammaln(df / 2)
        - 0.5 * jnp.log(df * jnp.pi)
        - 0.5 * (df + 1) * jnp.log1p(x**2 / df)
    )


def t_pdf(x, df) -> Array:
    return jnp.exp(t_logpdf(x, df))


def t_cdf(x, df) -> Array:
    return 0.5 * (
        1.0 + jnp.sign(x) - jnp.sign(x) * betainc(df / 2, 0.5, df / (df + x**2))
    )


def t_logcdf(x, df) -> Array:
    return jnp.log(t_cdf(x, df))


def t_sf(x, df) -> Array:
    return 1 - 0.5 * (
        1.0 + jnp.sign(x) - jnp.sign(x) * betainc(df / 2, 0.5, df / (df + x**2))
    )


def t_ppf(x, df) -> Array:
    beta_val = betaincinv(df / 2, 0.5, 1 - jnp.abs(2 * x - 1))
    return jnp.sqrt(df * (1 - beta_val) / beta_val) * jnp.sign(x - 0.5)





import jax.numpy as jnp
import jax.random as jr
import jax.tree_util as jtu
from jaxtyping import ArrayLike

from ...core import (
    _cdf_impl,
    _check_params_equal_tree_strcutre,
    _intialize_params_tree,
    _kurtosis_impl,
    _logcdf_impl,
    _logpdf_impl,
    _mean_impl,
    _params_impl,
    _pdf_impl,
    _quantile_impl,
    _rand_impl,
    _sf_impl,
    _skewness_impl,
    _standard_dev_impl,
    _support_impl,
    _variance_impl,
    DTypeLikeFloat,
    KeyArray,
    PyTreeVar,
    Shape,
)
from ...dist_math.beta import (
    beta_cdf,
    beta_logcdf,
    beta_logpdf,
    beta_pdf,
    beta_ppf,
    beta_sf,
)
from ...random_utils import split_tree
from ...tree_utils import tree_map_dist_at
from .._base import ContinuousUnivariateDistribution


class Beta(ContinuousUnivariateDistribution):
    """Beta distribution.

    Args:
        a (PyTreeVar): Shape parameter a.
        b (PyTreeVar): Shape parameter b.
        dtype (jax.numpy.dtype): dtype of the distribution, default jnp.float_.
        use_batch (bool): Whether to use with vmap. Default False.

    Examples:
        >>> import jax.numpy as jnp
        >>> from fenbux import Beta, logpdf
        >>> dist = Beta(1.0, 1.0)
        >>> logpdf(dist, jnp.ones((10, )))
    """

    a: PyTreeVar
    b: PyTreeVar

    def __init__(
        self, a: PyTreeVar = 0.0, b: PyTreeVar = 0.0, dtype=jnp.float_, use_batch=False
    ):
        _check_params_equal_tree_strcutre(a, b, use_batch=use_batch)
        self.a, self.b = _intialize_params_tree(a, b, use_batch=use_batch, dtype=dtype)


@_params_impl.dispatch
def _params_impl(dist: Beta):
    return dist.a, dist.b


@_support_impl.dispatch
def _support(d: Beta):
    dist = d.broadcast_params()
    return jtu.tree_map(lambda a: jnp.zeros_like(a), dist.a), jtu.tree_map(
        lambda a: jnp.ones_like(a), dist.a
    )


@_mean_impl.dispatch
def _mean(d: Beta):
    dist = d.broadcast_params()
    return jtu.tree_map(lambda a, b: a / (a + b), dist.a, dist.b)


@_variance_impl.dispatch
def _variance(d: Beta):
    dist = d.broadcast_params()
    return jtu.tree_map(
        lambda a, b: a * b / ((a + b) ** 2 * (a + b + 1)), dist.a, dist.b
    )


@_standard_dev_impl.dispatch
def _standard_dev(d: Beta):
    dist = d.broadcast_params()
    return jtu.tree_map(
        lambda a, b: jnp.sqrt(a * b / ((a + b) ** 2 * (a + b + 1))), dist.a, dist.b
    )


@_skewness_impl.dispatch
def _skewness(d: Beta):
    dist = d.broadcast_params()
    return jtu.tree_map(
        lambda a, b: 2
        * (b - a)
        * jnp.sqrt(a + b + 1)
        / ((a + b + 2) * jnp.sqrt(a * b)),
        dist.a,
        dist.b,
    )


@_kurtosis_impl.dispatch
def _kurtosis(d: Beta):
    dist = d.broadcast_params()
    return jtu.tree_map(
        lambda a, b: (
            6
            * ((a - b) ** 2 * (a + b + 1) - a * b * (a + b + 2))
            / (a * b * (a + b + 2) * (a + b + 3))
        ),
        dist.a,
        dist.b,
    )


@_logpdf_impl.dispatch
def _logpdf(d: Beta, x: ArrayLike):
    dist = d.broadcast_params()
    return tree_map_dist_at(beta_logpdf, dist, x)


@_pdf_impl.dispatch
def _pdf(d: Beta, x: ArrayLike):
    return tree_map_dist_at(beta_pdf, d, x)


@_logcdf_impl.dispatch
def _logcdf(d: Beta, x: ArrayLike):
    dist = d.broadcast_params()
    return tree_map_dist_at(beta_logcdf, dist, x)


@_cdf_impl.dispatch
def _cdf(d: Beta, x: ArrayLike):
    dist = d.broadcast_params()
    return tree_map_dist_at(beta_cdf, dist, x)


@_quantile_impl.dispatch
def _quantile(d: Beta, x: ArrayLike):
    dist = d.broadcast_params()
    return tree_map_dist_at(beta_ppf, dist, x)


@_sf_impl.dispatch
def _sf(d: Beta, x: ArrayLike):
    dist = d.broadcast_params()
    return tree_map_dist_at(beta_sf, dist, x)


@_rand_impl.dispatch
def _rand(d: Beta, key: KeyArray, shape: Shape = (), dtype: DTypeLikeFloat = float):
    dist = d.broadcast_params()
    _key_tree = split_tree(key, dist.a)
    return jtu.tree_map(
        lambda a, b, k: jr.beta(k, a, b, shape=shape, dtype=dtype),
        dist.a,
        dist.b,
        _key_tree,
    )



from jax import numpy as jnp
from jaxtyping import Array


def logistic_logpdf(x, loc, scale) -> Array:
    z = (x - loc) / scale
    return -jnp.log(scale) - z - 2 * jnp.log1p(jnp.exp(-z))


def logistic_pdf(x, loc, scale) -> Array:
    return jnp.exp(logistic_logpdf(x, loc, scale))


def logistic_cdf(x, loc, scale) -> Array:
    z = (x - loc) / scale
    return 1 / (1 + jnp.exp(-z))


def logistic_logcdf(x, loc, scale) -> Array:
    return -jnp.log1p(jnp.exp(-(x - loc) / scale))


def logistic_sf(x, loc, scale) -> Array:
    return 1 - logistic_cdf(x, loc, scale)


def logistic_logsf(x, loc, scale) -> Array:
    return -jnp.log(logistic_cdf(x, loc, scale))


def logistic_ppf(x, loc, scale) -> Array:
    return loc + scale * jnp.log(x / (1 - x))


def logistic_isf(x, loc, scale) -> Array:
    return logistic_ppf(1 - x, loc, scale)


def logistic_cf(t, loc, scale) -> Array:
    return jnp.exp(1j * t * loc) * (jnp.pi * scale * t) / jnp.sinh(jnp.pi * scale * t)



from jax import numpy as jnp
from jax.scipy.special import erf, erfinv
from jaxtyping import Array


def lognormal_logpdf(x, mean, sigma) -> Array:
    return (
        -0.5 * jnp.log(2 * jnp.pi)
        - jnp.log(sigma)
        - jnp.log(x)
        - ((jnp.log(x) - mean) / sigma) ** 2 / 2
    )


def lognormal_pdf(x, mean, sigma) -> Array:
    return jnp.exp(lognormal_logpdf(x, mean, sigma))


def lognormal_cdf(x, mean, sigma) -> Array:
    return 0.5 * (1 + erf((jnp.log(x) - mean) / (sigma * jnp.sqrt(2))))


def lognormal_logcdf(x, mean, sigma) -> Array:
    return jnp.log(0.5 * (1 + erf((jnp.log(x) - mean) / (sigma * jnp.sqrt(2)))))


def lognormal_sf(x, mean, sigma) -> Array:
    return 1 - lognormal_cdf(x, mean, sigma)


def lognormal_logsf(x, mean, sigma) -> Array:
    return jnp.log(lognormal_sf(x, mean, sigma))


def lognormal_ppf(x, mean, sigma) -> Array:
    return jnp.exp(mean + sigma * erfinv(2 * x - 1) * jnp.sqrt(2))


def lognormal_isf(x, mean, sigma) -> Array:
    return lognormal_ppf(1 - x, mean, sigma)



from jax import numpy as jnp
from jax.scipy.special import gammainc
from jax.scipy.stats.gamma import logpdf as _jax_gamma_logpdf
from jaxtyping import Array
from tensorflow_probability.substrates.jax.math import igammainv


def gamma_logpdf(x, shape, rate) -> Array:
    return _jax_gamma_logpdf(x, shape, scale=1 / rate)


def gamma_pdf(x, shape, rate) -> Array:
    return jnp.exp(gamma_logpdf(x, shape, rate))


def gamma_cdf(x, shape, rate) -> Array:
    return gammainc(shape, rate * x)


def gamma_logcdf(x, shape, rate) -> Array:
    return jnp.log(gamma_cdf(x, shape, rate))


def gamma_sf(x, shape, rate) -> Array:
    return 1 - gamma_cdf(x, shape, rate)


def gamma_logsf(x, shape, rate) -> Array:
    return jnp.log(gamma_sf(x, shape, rate))


def gamma_ppf(x, shape, rate) -> Array:
    return igammainv(shape, x) / rate


def gamma_isf(x, shape, rate) -> Array:
    return gamma_ppf(1 - x, shape, rate)


def gamma_mgf(t, shape, rate) -> Array:
    return (1 - rate * t) ** (-shape)


def gamma_cf(t, shape, rate) -> Array:
    return (1 - rate * 1j * t) ** (-shape)



import jax.numpy as jnp
import jax.random as jr
import jax.tree_util as jtu

from ...core import (
    _cdf_impl,
    _check_params_equal_tree_strcutre,
    _entropy_impl,
    _intialize_params_tree,
    _kurtosis_impl,
    _logcdf_impl,
    _logpdf_impl,
    _mean_impl,
    _params_impl,
    _pdf_impl,
    _quantile_impl,
    _rand_impl,
    _sf_impl,
    _skewness_impl,
    _standard_dev_impl,
    _support_impl,
    _variance_impl,
    DTypeLikeFloat,
    KeyArray,
    PyTreeVar,
    Shape,
)
from ...dist_math.pareto import (
    pareto_cdf,
    pareto_logcdf,
    pareto_logpdf,
    pareto_pdf,
    pareto_ppf,
    pareto_sf,
)
from ...random_utils import split_tree
from ...tree_utils import tree_map_dist_at
from .._base import ContinuousUnivariateDistribution


class Pareto(ContinuousUnivariateDistribution):
    """Pareto distribution.

        X ~ Pareto(shape, scale)

    Args:
        shape (PyTree): Shape parameter of the distribution.
        scale (PyTree): Scale parameter of the distribution.
        dtype (jax.numpy.dtype): dtype of the distribution, default jnp.float_.
        use_batch (bool): Whether to use with vmap. Default False.

    Examples:
        >>> import jax.numpy as jnp
        >>> from fenbux import Pareto, logpdf
        >>> dist = Pareto(1.0, 1.0)
        >>> logpdf(dist, jnp.ones((10, )))
    """

    shape: PyTreeVar
    scale: PyTreeVar

    def __init__(self, shape=0.0, scale=0.0, dtype=jnp.float_, use_batch=False):
        _check_params_equal_tree_strcutre(shape, scale, use_batch=use_batch)
        self.shape, self.scale = _intialize_params_tree(
            shape, scale, use_batch=use_batch, dtype=dtype
        )


@_params_impl.dispatch
def _params(d: Pareto):
    return (d.shape, d.scale)


@_support_impl.dispatch
def _support(d: Pareto):
    d = d.broadcast_params()
    return jtu.tree_map(lambda scale: scale, d.scale), jtu.tree_map(
        lambda scale: jnp.full_like(scale, jnp.inf), d.scale
    )


@_mean_impl.dispatch
def _mean(d: Pareto):
    d = d.broadcast_params()
    return jtu.tree_map(
        lambda a, b: jnp.where(a > 1.0, b * a / (a - 1.0), jnp.inf), d.shape, d.scale
    )


@_variance_impl.dispatch
def _variance(d: Pareto):
    d = d.broadcast_params()
    return jtu.tree_map(
        lambda a, b: jnp.where(
            a > 2.0, b**2 * a / ((a - 1.0) ** 2 * (a - 2.0)), jnp.inf
        ),
        d.shape,
        d.scale,
    )


@_standard_dev_impl.dispatch
def _standard_dev(d: Pareto):
    d = d.broadcast_params()
    return jtu.tree_map(
        lambda a, b: jnp.where(
            a > 2.0, b * jnp.sqrt(a) / ((a - 1.0) * jnp.sqrt(a - 2.0)), jnp.inf
        ),
        d.shape,
        d.scale,
    )


@_skewness_impl.dispatch
def _skewness(d: Pareto):
    d = d.broadcast_params()
    return jtu.tree_map(
        lambda a: jnp.where(
            a > 3, 2.0 * (1.0 + a) / (a - 3.0) * jnp.sqrt((a - 2.0) / a), jnp.nan
        ),
        d.shape,
    )


@_kurtosis_impl.dispatch
def _kurtosis(d: Pareto):
    d = d.broadcast_params()
    return jtu.tree_map(
        lambda a: jnp.where(
            a > 4,
            6.0 * (a**3 + a**2 - 6.0 * a - 2.0) / (a * (a - 3.0) * (a - 4.0)),
            jnp.nan,
        ),
        d.shape,
    )


@_entropy_impl.dispatch
def _entropy(d: Pareto):
    d = d.broadcast_params()
    return jtu.tree_map(
        lambda a, b: jnp.log(b) - jnp.log(a) + 1.0 / a + 1.0, d.shape, d.scale
    )


@_logpdf_impl.dispatch
def _logpdf(d: Pareto, x):
    d = d.broadcast_params()
    return tree_map_dist_at(pareto_logpdf, d, x)


@_pdf_impl.dispatch
def _pdf(d: Pareto, x):
    d = d.broadcast_params()
    return tree_map_dist_at(pareto_pdf, d, x)


@_logcdf_impl.dispatch
def _logcdf(d: Pareto, x):
    d = d.broadcast_params()
    return tree_map_dist_at(pareto_logcdf, d, x)


@_cdf_impl.dispatch
def _cdf(d: Pareto, x):
    d = d.broadcast_params()
    return tree_map_dist_at(pareto_cdf, d, x)


@_quantile_impl.dispatch
def _quantile(d: Pareto, x):
    d = d.broadcast_params()
    return tree_map_dist_at(pareto_ppf, d, x)


@_sf_impl.dispatch
def _sf(d: Pareto, x):
    d = d.broadcast_params()
    return tree_map_dist_at(pareto_sf, d, x)


@_rand_impl.dispatch
def _rand(d: Pareto, key: KeyArray, shape: Shape = (), dtype: DTypeLikeFloat = float):
    d = d.broadcast_params()
    _key_tree = split_tree(key, d.shape)
    return jtu.tree_map(
        lambda a, b, k: jr.pareto(k, a, shape, dtype) * b, d.shape, d.scale, _key_tree
    )



import jax.numpy as jnp
import jax.tree_util as jtu
from jaxtyping import ArrayLike

from ..core import AbstractDistribution
from ..core._abstract_impls import (
    _cdf_impl,
    _censor_impl,
    _logcdf_impl,
    _logpdf_impl,
    _logpmf_impl,
    _params_impl,
    _pdf_impl,
    _pmf_impl,
    _support_impl,
)
from ._base import (
    ContinuousUnivariateDistribution,
    DiscreteUnivariateDistribution,
    TransformedDistribution,
)


class CensoredDistribution(TransformedDistribution):
    lower: ArrayLike
    upper: ArrayLike
    uncensored: AbstractDistribution
    """A distribution censored between lower and upper.
    
    Args:
        lower (ArrayLike): Lower bound of the distribution.
        upper (ArrayLike): Upper bound of the distribution.
        d (AbstractDistribution): The uncensored distribution.
    """

    def __init__(self, lower, upper, d):
        self.lower = lower
        self.upper = upper
        self.uncensored = d


class ContinuousCensoredDistribution(CensoredDistribution):
    def __init__(self, lower, upper, d):
        super().__init__(lower, upper, d)


class DiscreteCensoredDistribution(CensoredDistribution):
    def __init__(self, lower, upper, d):
        super().__init__(lower, upper, d)


@_censor_impl.dispatch
def _censor_general(d: AbstractDistribution, lower, upper):
    return CensoredDistribution(lower, upper, d)


@_censor_impl.dispatch
def _censor(d: ContinuousUnivariateDistribution, lower, upper):
    return ContinuousCensoredDistribution(lower, upper, d)


@_censor_impl.dispatch
def _censor(d: DiscreteUnivariateDistribution, lower, upper):
    return DiscreteCensoredDistribution(lower, upper, d)


@_params_impl.dispatch
def _params(d: CensoredDistribution):
    return (d.lower, d.upper, _params_impl(d.uncensored))


@_pdf_impl.dispatch
def _pdf_general(d: CensoredDistribution, x: ArrayLike):
    _pdf = _pdf_impl(d.uncensored, x)
    supp = _support_impl(d.uncensored)
    _pdf = jtu.tree_map(
        lambda _p, _l: jnp.where(x < jnp.maximum(_l, d.lower), 0.0, _p),
        _pdf,
        supp[0],
    )
    _pdf = jtu.tree_map(
        lambda _p, _u: jnp.where(x > jnp.minimum(_u, d.upper), 0.0, _p),
        _pdf,
        supp[1],
    )
    return _pdf


@_pdf_impl.dispatch
def _pdf(d: ContinuousCensoredDistribution, x: ArrayLike):
    _pdf = _pdf_impl(d.uncensored, x)
    supp = _support_impl(d.uncensored)
    _pdf = jtu.tree_map(
        lambda _p, _l: jnp.where(x < jnp.maximum(_l, d.lower), 0.0, _p),
        _pdf,
        supp[0],
    )
    _pdf = jtu.tree_map(
        lambda _p, _u: jnp.where(x > jnp.minimum(_u, d.upper), 0.0, _p),
        _pdf,
        supp[1],
    )
    return _pdf


@_logpdf_impl.dispatch
def _logpdf_general(d: CensoredDistribution, x: ArrayLike):
    _lpdf = _logpdf_impl(d.uncensored, x)
    supp = _support_impl(d.uncensored)
    _lpdf = jtu.tree_map(
        lambda _lp, _l: jnp.where(
            x < jnp.where(x < jnp.maximum(_l, d.lower), 0.0, _lp),
            -jnp.inf,
            _lp,
        ),
        _lpdf,
        supp[0],
    )
    _lpdf = jtu.tree_map(
        lambda _lp, _u: jnp.where(x > jnp.minimum(_u, d.upper), 0.0, _lp),
        _lpdf,
        supp[1],
    )
    return _lpdf


@_logpdf_impl.dispatch
def _logpdf(d: ContinuousCensoredDistribution, x: ArrayLike):
    _lpdf = _logpdf_impl(d.uncensored, x)
    supp = _support_impl(d.uncensored)
    _lpdf = jtu.tree_map(
        lambda _lp, _l: jnp.where(
            x < jnp.where(x < jnp.maximum(_l, d.lower), 0.0, _lp),
            -jnp.inf,
            _lp,
        ),
        _lpdf,
        supp[0],
    )
    _lpdf = jtu.tree_map(
        lambda _lp, _u: jnp.where(x > jnp.minimum(_u, d.upper), 0.0, _lp),
        _lpdf,
        supp[1],
    )
    return _lpdf


@_cdf_impl.dispatch
def _cdf_general(d: CensoredDistribution, x: ArrayLike):
    _cdf = _cdf_impl(d.uncensored, x)
    supp = _support_impl(d.uncensored)
    _cdf = jtu.tree_map(
        lambda _c, _l: jnp.where(x < jnp.maximum(_l, d.lower), 0.0, _c),
        _cdf,
        supp[0],
    )
    _cdf = jtu.tree_map(
        lambda _c, _u: jnp.where(x > jnp.minimum(_u, d.upper), 1.0, _c),
        _cdf,
        supp[1],
    )
    return _cdf


@_cdf_impl.dispatch
def _cdf(d: ContinuousCensoredDistribution, x: ArrayLike):
    _cdf = _cdf_impl(d.uncensored, x)
    supp = _support_impl(d.uncensored)
    _cdf = jtu.tree_map(
        lambda _c, _l: jnp.where(x < jnp.maximum(_l, d.lower), 0.0, _c),
        _cdf,
        supp[0],
    )
    _cdf = jtu.tree_map(
        lambda _c, _u: jnp.where(x > jnp.minimum(_u, d.upper), 1.0, _c),
        _cdf,
        supp[1],
    )
    return _cdf


@_logcdf_impl.dispatch
def _logcdf(d: ContinuousCensoredDistribution, x: ArrayLike):
    _lcdf = _logcdf_impl(d.uncensored, x)
    supp = _support_impl(d.uncensored)
    _lcdf = jtu.tree_map(
        lambda _lc, _l: jnp.where(x < jnp.maximum(_l, d.lower), -jnp.inf, _lc),
        _lcdf,
        supp[0],
    )
    _lcdf = jtu.tree_map(
        lambda _lc, _u: jnp.where(x > jnp.minimum(_u, d.upper), 0.0, _lc),
        _lcdf,
        supp[1],
    )
    return _lcdf


@_pmf_impl.dispatch
def _pmf(d: DiscreteCensoredDistribution, x: ArrayLike):
    _pmf = _pmf_impl(d.uncensored, x)
    supp = _support_impl(d.uncensored)
    _pmf = jtu.tree_map(
        lambda _p, _l: jnp.where(x < jnp.maximum(_l, d.lower), 0.0, _p),
        _pmf,
        supp[0],
    )
    _pmf = jtu.tree_map(
        lambda _p, _u: jnp.where(x > jnp.minimum(_u, d.upper), 0.0, _p),
        _pmf,
        supp[1],
    )
    return _pmf


@_logpmf_impl.dispatch
def _logpmf(d: DiscreteCensoredDistribution, x: ArrayLike):
    _lpmf = _logpmf_impl(d.uncensored, x)
    supp = _support_impl(d.uncensored)
    _lpmf = jtu.tree_map(
        lambda _lp, _l: jnp.where(x < jnp.maximum(_l, d.lower), -jnp.inf, _lp),
        _lpmf,
        supp[0],
    )
    _lpmf = jtu.tree_map(
        lambda _lp, _u: jnp.where(x > jnp.minimum(_u, d.upper), 0.0, _lp),
        _lpmf,
        supp[1],
    )
    return _lpmf



import jax.numpy as jnp
import jax.random as jr
import jax.tree_util as jtu
from jaxtyping import ArrayLike

from ...core import (
    _cdf_impl,
    _cf_impl,
    _check_params_equal_tree_strcutre,
    _intialize_params_tree,
    _kurtosis_impl,
    _logcdf_impl,
    _logpmf_impl,
    _mean_impl,
    _mgf_impl,
    _params_impl,
    _pmf_impl,
    _quantile_impl,
    _rand_impl,
    _sf_impl,
    _skewness_impl,
    _standard_dev_impl,
    _support_impl,
    _variance_impl,
    DTypeLikeFloat,
    KeyArray,
    PyTreeVar,
    Shape,
)
from ...dist_math.binomial import (
    binom_cdf,
    binom_cf,
    binom_logcdf,
    binom_logpmf,
    binom_mgf,
    binom_pmf,
    binom_ppf,
    binom_sf,
)
from ...random_utils import split_tree
from ...tree_utils import tree_map_dist_at
from .._base import DiscreteUnivariateDistribution


class Binomial(DiscreteUnivariateDistribution):
    """Binomial distribution.

            X ~ Binomial(n, p)

    Args:
        n (PyTree): Number of trials.
        p (PyTree): Probability of success.
        dtype (jax.numpy.dtype): dtype of the distribution, default jnp.float_.
        use_batch (bool): Whether to use with vmap. Default False.

    Examples:
        >>> import jax.numpy as jnp
        >>> from fenbux import Binomial, logpmf
        >>> dist = Binomial(10, 0.5)
        >>> logpmf(dist, jnp.ones((10, )))
    """

    n: PyTreeVar
    p: PyTreeVar

    def __init__(self, n=0.0, p=0.0, dtype=jnp.float_, use_batch=False):
        _check_params_equal_tree_strcutre(n, p, use_batch=use_batch)
        self.n, self.p = _intialize_params_tree(n, p, use_batch=use_batch, dtype=dtype)


@_params_impl.dispatch
def _params(d: Binomial):
    return (d.n, d.p)


@_support_impl.dispatch
def _support(d: Binomial):
    d = d.broadcast_params()
    return jtu.tree_map(lambda n: jnp.zeros_like(n), d.n), jtu.tree_map(
        lambda n: n, d.n
    )


@_mean_impl.dispatch
def _mean(d: Binomial):
    d = d.broadcast_params()
    return jtu.tree_map(lambda p, n: p * n, d.p, d.n)


@_variance_impl.dispatch
def _variance(d: Binomial):
    d = d.broadcast_params()
    return jtu.tree_map(lambda p, n: p * (1 - p) * n, d.p, d.n)


@_standard_dev_impl.dispatch
def _standard_dev(d: Binomial):
    d = d.broadcast_params()
    return jtu.tree_map(lambda p, n: jnp.sqrt(p * (1 - p) * n), d.p, d.n)


@_skewness_impl.dispatch
def _skewness(d: Binomial):
    d = d.broadcast_params()
    return jtu.tree_map(lambda p, n: (1 - 2 * p) / jnp.sqrt(p * (1 - p) * n), d.p, d.n)


@_kurtosis_impl.dispatch
def _kurtosis(d: Binomial):
    d = d.broadcast_params()
    return jtu.tree_map(
        lambda p, n: (1 - 6 * p * (1 - p)) / (p * (1 - p) * n), d.p, d.n
    )


@_logpmf_impl.dispatch
def _logpmf(d: Binomial, x: ArrayLike):
    d = d.broadcast_params()
    return tree_map_dist_at(binom_logpmf, d, x)


@_pmf_impl.dispatch
def _pmf(d: Binomial, x: ArrayLike):
    d = d.broadcast_params()
    return tree_map_dist_at(binom_pmf, d, x)


@_logcdf_impl.dispatch
def _logcdf(d: Binomial, x: ArrayLike):
    d = d.broadcast_params()
    return tree_map_dist_at(binom_logcdf, d, x)


@_cdf_impl.dispatch
def _cdf(d: Binomial, x: ArrayLike):
    d = d.broadcast_params()
    return tree_map_dist_at(binom_cdf, d, x)


@_quantile_impl.dispatch
def _quantile(d: Binomial, q: ArrayLike):
    d = d.broadcast_params()
    return tree_map_dist_at(binom_ppf, d, q)


@_mgf_impl.dispatch
def _mgf(d: Binomial, t: ArrayLike):
    d = d.broadcast_params()
    return tree_map_dist_at(binom_mgf, d, t)


@_cf_impl.dispatch
def _cf(d: Binomial, t: ArrayLike):
    d = d.broadcast_params()
    return tree_map_dist_at(binom_cf, d, t)


@_sf_impl.dispatch
def _sf(d: Binomial, x: ArrayLike):
    d = d.broadcast_params()
    return tree_map_dist_at(binom_sf, d, x)


@_rand_impl.dispatch
def _rand(d: Binomial, key: KeyArray, shape: Shape = (), dtype: DTypeLikeFloat = float):
    _dist = d.broadcast_params()
    _key_tree = split_tree(key, _dist.n)
    rvs = jtu.tree_map(
        lambda p, n, k: jr.binomial(k, n, p, shape=shape, dtype=dtype),
        _dist.p,
        _dist.n,
        _key_tree,
    )
    return rvs



from . import (
    bernoulli,
    beta,
    binomial,
    chi2,
    exp,
    f,
    gamma,
    logistic,
    lognormal,
    normal,
    pareto,
    poisson,
    t,
    uniform,
    wald,
    weibull,
)



from jax import numpy as jnp
from jax.scipy.special import betainc, betaln, xlogy
from jaxtyping import Array

from ..extension import fdtri


def f_logpdf(x, dfn, dfd) -> Array:
    return (
        dfd / 2 * jnp.log(dfd)
        + dfn / 2 * jnp.log(dfn)
        + xlogy(dfn / 2 - 1, x)
        - (((dfn + dfd) / 2) * jnp.log(dfd + dfn * x) + betaln(dfn / 2, dfd / 2))
    )


def f_pdf(x, dfn, dfd) -> Array:
    return jnp.exp(f_logpdf(x, dfn, dfd))


def f_cdf(x, dfn, dfd) -> Array:
    return betainc(dfn / 2, dfd / 2, dfn * x / (dfd + dfn * x))


def f_logcdf(x, dfn, dfd) -> Array:
    return jnp.log(betainc(dfn / 2, dfd / 2, dfn * x / (dfn * x + dfd)))


def f_sf(x, dfn, dfd) -> Array:
    return 1 - betainc(dfn / 2, dfd / 2, dfn * x / (dfd + dfn * x))


def f_logsf(x, dfn, dfd) -> Array:
    return jnp.log(f_sf(x, dfn, dfd))


def f_ppf(x, dfn, dfd) -> Array:
    return fdtri(dfn, dfd, x)


def f_isf(x, dfn, dfd) -> Array:
    return fdtri(dfd, dfn, 1 - x)



from jax import numpy as jnp
from jaxtyping import Array


def uniform_logpdf(x, lower, upper) -> Array:
    return jnp.where(
        (x < lower) | (x > upper),
        -jnp.inf,
        -jnp.log(upper - lower),
    )


def uniform_pdf(x, lower, upper) -> Array:
    return jnp.where(
        (x < lower) | (x > upper),
        0,
        1 / (upper - lower),
    )


def uniform_cdf(x, lower, upper) -> Array:
    return jnp.where(
        x < lower,
        0,
        jnp.where(
            x > upper,
            1,
            (x - lower) / (upper - lower),
        ),
    )


def uniform_logcdf(x, lower, upper) -> Array:
    return jnp.where(
        x < lower,
        -jnp.inf,
        jnp.where(
            x > upper,
            0,
            jnp.log((x - lower) / (upper - lower)),
        ),
    )


def uniform_sf(x, lower, upper) -> Array:
    return jnp.where(
        x < lower,
        1,
        jnp.where(
            x > upper,
            0,
            (upper - x) / (upper - lower),
        ),
    )


def uniform_logsf(x, lower, upper) -> Array:
    return jnp.where(
        x < lower,
        0,
        jnp.where(
            x > upper,
            -jnp.inf,
            jnp.log((upper - x) / (upper - lower)),
        ),
    )


def uniform_ppf(x, lower, upper) -> Array:
    return jnp.where(
        (x < 0) | (x > 1),
        jnp.nan,
        lower + x * (upper - lower),
    )


def uniform_isf(x, lower, upper) -> Array:
    return jnp.where(
        (x < 0) | (x > 1),
        jnp.nan,
        upper - x * (upper - lower),
    )


def uniform_mgf(t, lower, upper) -> Array:
    return jnp.where(
        t == 0,
        1,
        (jnp.exp(t * upper) - jnp.exp(t * lower)) / (t * (upper - lower)),
    )


def uniform_cf(t, lower, upper) -> Array:
    return jnp.where(
        t == 0,
        1,
        (jnp.exp(1j * t * upper) - jnp.exp(1j * t * lower))
        / (1j * t * (upper - lower)),
    )



from typing import Callable

import jax.random as jr
import jax.tree_util as jtu
from jaxtyping import PyTree

from .core import KeyArray, PyTreeKey


def split_tree(key: KeyArray, pytree: PyTree, *, is_leaf: Callable = None) -> PyTreeKey:
    """Split a key into a pytree of keys.

    Args:
        key (KeyArray): A jax.random.KeyArray.
        pytree (PyTree): A pytree to split.
        is_leaf (Callable): A function to determine whether a pytree node is a leaf. Default is jtu.is_leaf.

    Returns:
        PyTreeKey: A pytree of keys.
    """
    _, treedef = jtu.tree_flatten(pytree, is_leaf=is_leaf)
    keys = jr.split(key, treedef.num_leaves)
    return jtu.tree_unflatten(treedef, keys)



from ._beta import Beta
from ._chisquare import Chisquare
from ._exponential import Exponential
from ._f import F
from ._gamma import Gamma
from ._logistic import Logistic
from ._lognormal import LogNormal
from ._normal import Normal
from ._pareto import Pareto
from ._t import StudentT
from ._uniform import Uniform
from ._wald import Wald
from ._weibull import Weibull



from .core._dist import AbstractDistribution, ParamShape
from .core._typing import (
    ArrayLike,
    DTypeLikeFloat,
    DTypeLikeInt,
    KeyArray,
    PyTreeKey,
    PyTreeVar,
    Shape,
)
from .multivariate import (
    AbstractMultivariateDistribution,
    ContinuousMultivariateDistribution,
)
from .scipy_stats import ScipyDist
from .univariate import (
    AffineDistribution,
    ContinuousAffineDistribution,
    ContinuousUnivariateDistribution,
    DiscreteAffineDistribution,
    DiscreteUnivariateDistribution,
    UnivariateDistribution,
)



import numpy as np

import fenbux.tree_utils as fbtu


def f(x, y, z):
    return x + y + z


def g(a, b):
    return a * b


def test_broadcast_pytree_arrays_shape():
    arr1 = np.ones((2, 3))
    arr2 = np.ones(())

    assert fbtu.broadcast_pytree_arrays_shapes(arr1, arr2).shape == (2, 3)


def test_broadcast_pytree_arrays():
    arr1 = np.ones((2, 3))
    arr2 = np.ones(())

    assert fbtu.broadcast_pytree_arrays(arr1, arr2)[0].shape == (2, 3)
    assert fbtu.broadcast_pytree_arrays(arr1, arr2)[1].shape == (2, 3)


def test_tree_structures_all_eq():
    assert fbtu.tree_structures_all_eq(1, 2, 3)
    assert fbtu.tree_structures_all_eq((1, 2), (3, 4), (5, 6))
    assert not fbtu.tree_structures_all_eq((1, 2), (3, 4), (5, 6, 7))


def test_tree_map_with_kwargs():
    assert fbtu.tree_map(f, 1, 2, z=3) == 6
    assert fbtu.tree_map(f, (1, 2), (3, 4), z=(5, 6)) == (9, 12)


def test_tree_map_with_flat_kwargs():
    assert fbtu.tree_map(f, (1, 2), (3, 4), z=3, flat_kwargnames=("z",)) == (7, 9)
    assert fbtu.tree_map(f, (1, 2), y=2, z=3, flat_kwargnames=("z", "y")) == (6, 7)
    assert fbtu.tree_map(f, x=1, y=2, z=3, flat_kwargnames=("x", "y", "z")) == 6


def test_tree_map_of_tree_map():
    assert fbtu.tree_map(
        g, (1, 2), fbtu.tree_map(f, (1, 2), (3, 4), z=3, flat_kwargnames=("z",))
    ) == (7, 18)
    assert fbtu.tree_map(
        g, (1, 2), b=fbtu.tree_map(f, (1, 2), (3, 4), z=3, flat_kwargnames=("z",))
    ) == (7, 18)
    assert fbtu.tree_map(
        g, a=(1, 2), b=fbtu.tree_map(f, (1, 2), (3, 4), z=3, flat_kwargnames=("z",))
    ) == (7, 18)
    assert fbtu.tree_map(
        g,
        a=1,
        b=fbtu.tree_map(f, (1, 2), (3, 4), z=3, flat_kwargnames=("z",)),
        flat_kwargnames=("a"),
    ) == (7, 9)



import jax.numpy as jnp
from jaxtyping import ArrayLike

from ._abstract_impls import _affine_impl, _censor_impl, _truncate_impl
from ._dist import AbstractDistribution
from ._typing import PyTree


def affine(
    d: AbstractDistribution, loc: ArrayLike = 0.0, scale: ArrayLike = 1.0
) -> PyTree:
    """Affine transformation of a distribution
        y = loc + scale * x
    Args:
        d (AbstractDistribution): A distribution object.
        loc (ArrayLike): loc parameter of the affine transformation.
        scale (ArrayLike): scale parameter of the affine transformation.

    Example:
        >>> from fenbux import affine, logpdf
        >>> from fenbux.univariate import Normal
        >>> dist = Normal(0.0, 1.0)
        >>> aff_dist = affine(dist, 0.0, 1.0)
        >>> logpdf(aff_dist, 0.0)
    """
    return _affine_impl(d, loc, scale)


def truncate(
    d: AbstractDistribution, lower: ArrayLike = -jnp.inf, upper: ArrayLike = jnp.inf
) -> PyTree:
    """Truncate a distribution to a given interval.

    Args:
        d (AbstractDistribution): A distribution object.
        lower (ArrayLike): Lower bound of the truncated distribution.
        upper (ArrayLike): Upper bound of the truncated distribution.

    Example:
        >>> from fenbux import truncate, logpdf
        >>> from fenbux.univariate import Normal
        >>> dist = Normal(0.0, 1.0)
        >>> truncate(dist, -1.0, 1.0)
        >>> logpdf(dist, -2.0)

    """
    return _truncate_impl(d, lower, upper)


def censor(
    d: AbstractDistribution, lower: ArrayLike = -jnp.inf, upper: ArrayLike = jnp.inf
) -> PyTree:
    """Censor a distribution to a given interval.

    Args:
        d (AbstractDistribution): A distribution object.
        lower (ArrayLike): Lower bound of the censored distribution.
        upper (ArrayLike): Upper bound of the censored distribution.

    Example:
        >>> from fenbux import censor, logpdf
        >>> from fenbux.univariate import Normal
        >>> dist = Normal(0.0, 1.0)
        >>> censor(dist, -1.0, 1.0)
        >>> logpdf(dist, -2.0)

    """
    return _censor_impl(d, lower, upper)



from jax import numpy as jnp
from jax.scipy.special import betainc, betaln, xlog1py, xlogy
from jaxtyping import Array
from tensorflow_probability.substrates.jax.math import betaincinv


def beta_logpdf(x, a, b) -> Array:
    lp = xlog1py(b - 1.0, -x) + xlogy(a - 1.0, x)
    lp -= betaln(a, b)
    return lp


def beta_pdf(x, a, b) -> Array:
    return jnp.exp(beta_logpdf(x, a, b))


def beta_logcdf(x, a, b) -> Array:
    return jnp.log(betainc(a, b, x))


def beta_cdf(x, a, b) -> Array:
    return betainc(a, b, x)


def beta_ppf(p, a, b) -> Array:
    return betaincinv(a, b, p)


def beta_sf(x, a, b) -> Array:
    return betainc(a, b, 1 - x)


def beta_isf(x, a, b) -> Array:
    return betaincinv(a, b, 1 - x)


def beta_logsf(x, a, b) -> Array:
    return jnp.log(betainc(a, b, 1 - x))



import jax.numpy as jnp
import jax.random as jr
import jax.tree_util as jtu
from jax import lax
from jaxtyping import ArrayLike

from ...core import (
    _cdf_impl,
    _cf_impl,
    _entropy_impl,
    _intialize_params_tree,
    _kurtosis_impl,
    _logcdf_impl,
    _logpmf_impl,
    _mean_impl,
    _mgf_impl,
    _params_impl,
    _pmf_impl,
    _quantile_impl,
    _rand_impl,
    _sf_impl,
    _skewness_impl,
    _standard_dev_impl,
    _support_impl,
    _variance_impl,
    DTypeLikeFloat,
    KeyArray,
    PyTreeVar,
    Shape,
)
from ...dist_math.bernoulli import (
    bernoulli_cdf,
    bernoulli_cf,
    bernoulli_logcdf,
    bernoulli_logpmf,
    bernoulli_mgf,
    bernoulli_pmf,
    bernoulli_ppf,
    bernoulli_sf,
)
from ...random_utils import split_tree
from ...tree_utils import tree_map_dist_at
from .._base import DiscreteUnivariateDistribution


class Bernoulli(DiscreteUnivariateDistribution):
    """Bernoulli distribution.
        X ~ Bernoulli(p)
    Args:
        p (PyTree): Probability of success.
        dtype (jax.numpy.dtype): dtype of the distribution, default jnp.float_.
        use_batch (bool): Whether to use with vmap. Default False.

    Examples:
        >>> import jax.numpy as jnp
        >>> from fenbux import Bernoulli, logpdf
        >>> dist = Bernoulli(0.5)
        >>> logpdf(dist, jnp.ones((10, )))
    """

    p: PyTreeVar

    def __init__(self, p=0.0, dtype=jnp.float_, use_batch=False):
        self.p = _intialize_params_tree(p, use_batch=use_batch, dtype=dtype)


@_params_impl.dispatch
def _params(d: Bernoulli):
    return (d.p,)


@_support_impl.dispatch
def _support(d: Bernoulli):
    return jtu.tree_map(lambda p: jnp.zeros_like(p), d.p), jtu.tree_map(
        lambda p: jnp.ones_like(p), d.p
    )


@_mean_impl.dispatch
def _mean(d: Bernoulli):
    return jtu.tree_map(lambda p: p, d.p)


@_variance_impl.dispatch
def _variance(d: Bernoulli):
    return jtu.tree_map(lambda p: p * (1 - p), d.p)


@_standard_dev_impl.dispatch
def _standard_dev(d: Bernoulli):
    return jtu.tree_map(lambda p: jnp.sqrt(p * (1 - p)), d.p)


@_kurtosis_impl.dispatch
def _kurtosis(d: Bernoulli):
    return jtu.tree_map(
        lambda p: jnp.where(
            lax.gt(p, 0.0) & lax.lt(p, 1.0),
            (1 - 6 * p * (1 - p)) / (p * (1 - p)),
            jnp.nan,
        ),
        d.p,
    )


@_skewness_impl.dispatch
def _skewness(d: Bernoulli):
    return jtu.tree_map(
        lambda p: jnp.where(
            lax.gt(p, 0.0) & lax.lt(p, 1.0),
            (1 - 2 * p) / jnp.sqrt(p * (1 - p)),
            jnp.nan,
        ),
        d.p,
    )


@_entropy_impl.dispatch
def _entropy(d: Bernoulli):
    return jtu.tree_map(lambda p: -p * jnp.log(p) - (1 - p) * jnp.log(1 - p), d.p)


@_pmf_impl.dispatch
def _pmf(d: Bernoulli, x: ArrayLike):
    return tree_map_dist_at(bernoulli_pmf, d, x)


@_logpmf_impl.dispatch
def _logpmf(d: Bernoulli, x: ArrayLike):
    return tree_map_dist_at(bernoulli_logpmf, d, x)


@_rand_impl.dispatch
def _rand(
    d: Bernoulli, key: KeyArray, shape: Shape = (), dtype: DTypeLikeFloat = float
):
    _key_tree = split_tree(key, d.p)
    rvs = jtu.tree_map(
        lambda p, k: jr.bernoulli(k, p, shape=shape).astype(dtype),
        d.p,
        _key_tree,
    )
    return rvs


@_cdf_impl.dispatch
def _cdf(d: Bernoulli, x: ArrayLike):
    return tree_map_dist_at(bernoulli_cdf, d, x)


@_logcdf_impl.dispatch
def _logcdf(d: Bernoulli, x: ArrayLike):
    return tree_map_dist_at(bernoulli_logcdf, d, x)


@_quantile_impl.dispatch
def _quantile(d: Bernoulli, x: ArrayLike):
    return tree_map_dist_at(bernoulli_ppf, d, x)


@_mgf_impl.dispatch
def _mgf(d: Bernoulli, t: ArrayLike):
    return tree_map_dist_at(bernoulli_mgf, d, t)


@_cf_impl.dispatch
def _cf(d: Bernoulli, t: ArrayLike):
    return tree_map_dist_at(bernoulli_cf, d, t)


@_sf_impl.dispatch
def _sf(d: Bernoulli, x: ArrayLike):
    return tree_map_dist_at(bernoulli_sf, d, x)



from typing import Tuple

from jaxtyping import ArrayLike, PyTree
from plum import Dispatcher

from ._dist import AbstractDistribution
from ._typing import DTypeLike, KeyArray, Shape


_fenbux_dispatch = Dispatcher()


@_fenbux_dispatch.abstract
def _params_impl(dist: AbstractDistribution) -> Tuple[PyTree, ...]:
    ...


@_fenbux_dispatch.abstract
def _support_impl(dist: AbstractDistribution) -> Tuple[PyTree, ...]:
    ...


@_fenbux_dispatch.abstract
def _mean_impl(dist: AbstractDistribution) -> PyTree:
    ...


@_fenbux_dispatch.abstract
def _variance_impl(dist: AbstractDistribution) -> PyTree:
    ...


@_fenbux_dispatch.abstract
def _standard_dev_impl(dist: AbstractDistribution) -> PyTree:
    ...


@_fenbux_dispatch.abstract
def _skewness_impl(dist: AbstractDistribution) -> PyTree:
    ...


@_fenbux_dispatch.abstract
def _kurtosis_impl(dist: AbstractDistribution) -> PyTree:
    ...


@_fenbux_dispatch.abstract
def _entropy_impl(dist: AbstractDistribution) -> PyTree:
    ...


@_fenbux_dispatch.abstract
def _pdf_impl(dist: AbstractDistribution, x: ArrayLike) -> PyTree:
    ...


@_fenbux_dispatch.abstract
def _logpdf_impl(dist: AbstractDistribution, x: ArrayLike) -> PyTree:
    ...


@_fenbux_dispatch.abstract
def _logcdf_impl(dist: AbstractDistribution, x: ArrayLike) -> PyTree:
    ...


@_fenbux_dispatch.abstract
def _cdf_impl(dist: AbstractDistribution, x: ArrayLike) -> PyTree:
    ...


@_fenbux_dispatch.abstract
def _rand_impl(
    dist: AbstractDistribution,
    key: KeyArray,
    shape: Shape,
    dtype: DTypeLike,
) -> PyTree:
    ...


@_fenbux_dispatch.abstract
def _quantile_impl(dist: AbstractDistribution, p: ArrayLike) -> PyTree:
    ...


@_fenbux_dispatch.abstract
def _logpmf_impl(dist: AbstractDistribution, x: ArrayLike) -> PyTree:
    ...


@_fenbux_dispatch.abstract
def _pmf_impl(dist: AbstractDistribution, x: ArrayLike) -> PyTree:
    ...


@_fenbux_dispatch.abstract
def _mgf_impl(dist: AbstractDistribution, t: ArrayLike) -> PyTree:
    ...


@_fenbux_dispatch.abstract
def _sf_impl(dist: AbstractDistribution, x: ArrayLike) -> PyTree:
    ...


@_fenbux_dispatch.abstract
def _cf_impl(dist: AbstractDistribution, t: ArrayLike) -> PyTree:
    ...


@_fenbux_dispatch.abstract
def _affine_impl(d: AbstractDistribution, loc: ArrayLike, scale: ArrayLike) -> PyTree:
    ...


@_fenbux_dispatch.abstract
def _truncate_impl(
    d: AbstractDistribution, lower: ArrayLike, upper: ArrayLike
) -> AbstractDistribution:
    ...


@_fenbux_dispatch.abstract
def _censor_impl(
    d: AbstractDistribution, lower: ArrayLike, upper: ArrayLike
) -> AbstractDistribution:
    ...
